from typing import List


#TODO: Check how much everything breaks if I try to sort by token on pre-processing errors.

class mfd_syntax_err(SyntaxError):
    lineno:int
    errtype:str
    inst_terminal:bool
    late_terminal:bool

    # adding the late_terminal flag to mark whether or not the parser should stop before reaching pydantic.
    # this may be not necessary, but it will be easier to refactor to remove it later than to refactor to add it.
    def __init__(self, message:str, lineno:int, errtype:str, inst_terminal:bool, late_terminal:bool):
        super().__init__(message)
        self.lineno = lineno
        self.errtype = errtype
        self.inst_terminal = inst_terminal
        self.late_terminal = late_terminal

    def __str__(self):
        return f"{self.errtype} on line {self.lineno}: {self.args[0]}"

class mfd_unexpected_macro(mfd_syntax_err):
    def __init__(self, lineno:int, macro_name:str) :
        message = f"Unexpected macro '{macro_name}' on line {lineno}."
        super().__init__(message, lineno, "unexpected_macro", False, True)

class mfd_duplicated_macro(mfd_syntax_err):
    def __init__(self, instance_1:int, instance_2:int, macro_name:str) :
        message = f"Duplicated macro '{macro_name}' on lines {instance_1} and {instance_2}."
        super().__init__(message, instance_2, "duplicated_macro", False, True)

class mfd_remaining_template(mfd_syntax_err):
    def __init__(self, lineno:int):
        message = f"Remaining template marker on line {lineno}."
        super().__init__(message, lineno, "remaining_template", False, True)

class mfd_no_desc(mfd_syntax_err):
    def __init__(self, lineno:int, major_token:str):
        message = f"Novel token @{major_token} started without a description on line {lineno}."
        super().__init__(message, lineno, "no_desc", False, True)

class mfd_wrong_macro_token(mfd_syntax_err) :
    def __init__(self, lineno:int) :
        message = f"Wrong token used to define a macro on line {lineno}. Please use `@, not '@, when defining a macro."
        super().__init__(message, lineno, "wrong_macro_token", False, True)

class mfd_err:
    def __init__(self, line: int, errtype: str, token_context: List[str],
                token_name: str, msg: str) :
        self.line = line
        self.errtype = errtype

        if self.errtype != "macro_misuse" :
            self.token_context = token_context
            token_string = "@" + token_context[0]
            if len(token_context) > 1:
                for token in token_context[1:-1]:
                    token_string = token_string + "_" + token
                token_string = token_string + "-" + token_context[-1]
            self.token_string = token_string

            self.token_name = token_name
        else :
            self.token_string = "NA"
            self.token_name = "NA"

        self.msg = msg

    def __str__(self) :
        if self.errtype == "missing_field" :
            return f"Line {self.line: <10} : {self.token_string} is missing the field {self.token_name}."
        elif self.errtype == "incomplete_data" :
            return f"Line {self.line: <10} : {self.token_string} has incomplete information: {self.msg}."
        elif self.errtype == "macro_misuse" :
            return f"Line {self.line: <10} : {self.msg}"
        else:
            return f"Line {self.line: <10} : {self.token_string} is of the wrong type: {self.msg}."

class error_mngr:
    has_major_parsing = False
    # mode : ("ALL", "FIRST")
    # order: ("TYPE", "TOKENS", "LINE")
    def __init__(self, mode, order):
        self._syntax_err_coll = {}
        self._error_collection = {}
        self.has_major_parsing = False
        self.mode = mode
        self.order = order

    def add_syntax_err(self, err_obj:mfd_syntax_err):
        if err_obj.late_terminal :
            self.has_major_parsing = True
        if err_obj.inst_terminal :
            raise err_obj
        # TODO: Add ordering functionality, for now I'm just assuming lineno ordering
        if err_obj.lineno not in self._syntax_err_coll.keys() :
            self._syntax_err_coll[err_obj.lineno] = [err_obj]
        else :
            self._syntax_err_coll[err_obj.lineno].append(err_obj)

    def print_syntax_errors(self):
        print("\nSyntax errors:")
        for line in sorted(self._syntax_err_coll.keys()) :
            for err in self._syntax_err_coll[line] :
                print("\t" + str(err))

    def add_error(self, error_obj: mfd_err):
        if self.mode == "FIRST" :
            print("Stopping on first error: ")
            print("\t" + str(error_obj))
            raise SystemExit(0)

        if self.order == "TYPE" :
            keyval = error_obj.errtype
        elif self.order == "TOKENS" :
            keyval = error_obj.token_string
        elif self.order == "LINE" :
            keyval = error_obj.line

        if keyval not in self._error_collection.keys() :
            self._error_collection[keyval] = [error_obj]
        else :
            self._error_collection[keyval].append(error_obj)

    def print_errors(self):
        if self.order == "TYPE" :
            title = "Errors of type "
        elif self.order == "TOKENS" :
            title = "Errors on token "
        elif self.order == "LINE" :
            title = "Errors on line "

        for keyval in self._error_collection.keys() :
            if self.order != "LINE" :
                print(title + str(keyval) + ": ")
            for err in self._error_collection[keyval] :
                print("\t" + str(err))


from typing import List, Optional, Union, TypeVar, Tuple
import datetime
from pydantic import BaseModel as PydanticBaseModel
from pydantic import AnyUrl, root_validator

# bcodmo? + bag
# export to rtf?
# error on unknown fields
# data1 compliance ('dublin core'?)

# Validation of user-defined fields: in the future?
#       runtime clas creation?
#       parser level validation?

# Test against all data types;
#       Return which ones *can* submit to & which *cannot*
#       Can then re-run on specific mode to see specific errors

T = TypeVar('T')
DatumT = Tuple[int, T]
DataT = List[DatumT[T]]
OptDataT = Optional[DataT[T]]

class BaseModel(PydanticBaseModel) :
    class Config:
        extra = 'allow'

class IncompleteDataError(ValueError):
    code = "incomplete_data_error"

all_versions = ["1.0"]
################################
# Helper Models                #
################################
class StrDescModel(BaseModel):
    desc: DataT[str] #TODO: Find a way to make an exception?
    Note: OptDataT[str]

################################
# Field Models                 #
################################
class MEDFORDmodel(BaseModel) :
    desc: OptDataT[str]
    Version: DataT[str]
    @root_validator
    def check_version(cls, values) :
        if values['Version'] == [] :
            raise IncompleteDataError()

        actual_value = values['Version'][0][1]
        if actual_value not in all_versions :
            raise ValueError(f"Version {actual_value} is not a valid version.")
        
        return values
    
class Journal(StrDescModel):
    Volume: DataT[int]
    Issue: DataT[int]
    Pages: OptDataT[str] #TODO: Validation?

class Date(BaseModel):
    desc: Union[DataT[datetime.date], DataT[datetime.datetime]]
    Note: DataT[str]
    #changed type to note because type is a reserved keyword

class Contributor(StrDescModel) :
    ORCID: OptDataT[int]
    Assocation: OptDataT[str]
    Role: OptDataT[str]
    Email: OptDataT[str] #TODO: Email validation

    @root_validator
    def check_corresponding_has_contact(cls, v) :
        if v['Role'] is not None:
            roles = [r[1] for r in v['Role']]
            if "Corresponding Author" in roles and v["Email"] is None :
                raise IncompleteDataError("Corresponding Authors must have a provided validated email")
        return v

class Funding(StrDescModel) :
    ID: OptDataT[str] #TODO: Funding ID validation?

class Keyword(StrDescModel):
    pass

class Species(StrDescModel):
    Loc: DataT[str]
    ReefCollection: DataT[str] # TODO: Change to date with note?
    Cultured: DataT[str]
    CultureCollection: DataT[str]

class Method(StrDescModel):
    Type: DataT[str]
    Company: OptDataT[str]
    Sample: OptDataT[str]

class Project(StrDescModel):
    pass

class Expedition(StrDescModel):
    ShipName: OptDataT[str]
    CruiseID: OptDataT[str]
    MooringID: OptDataT[str]
    DiveNumber: OptDataT[int]
    Synonyms: OptDataT[str]

    #def check_at_least_one_identifier(cls, v) :
    #    values = {key:value for key, value in v[0].__dict__.items() if not key.startswith('__') and not callable(key)}
    #    has_shipname = values['ShipName'] is not None and values['CruiseID'] is not None
    #    has_mooring = values['MooringID'] is not None 
    #    has_divenumber = values['DiveNumber'] is not None 
    #    if not any([has_shipname, has_mooring, has_divenumber]) :
    #        raise ValueError('BCO-DMO requires at least one cruise identifier. Your choices are: \n' +
    #                            '     ShipName and CruiseID, MooringID, or DiveNumber.')
    #    return v

class ArbitraryFile(StrDescModel):
    #todo: change to filename
    Path: OptDataT[str]
    Destination: OptDataT[str]
    URI: OptDataT[AnyUrl]
    outpath: str = ""
    @root_validator
    def check_singular_path_subdirectory(cls, values):
        # TODO: Don't allow remote files, yet... Separate tag? RemoteFile?
        if (values['Path'] != None and len(values['Path']) > 1) :
            raise ValueError("Please create a separate @File tag for each recorded file.")
        if (values['Destination'] != None and len(values['Destination']) > 1):
            raise ValueError("MEDFORD does not currently support copying a file multiple times through one tag. " + 
                            "Please use a separate @File tag for each output file.")
        #v = create_new_bagit_loc(v, "local")
        return values
    
    @root_validator
    def check_path_or_uri(cls, values) :
        if (values['Path'] == None or len(values['Path']) == 0) and (values['URI'] == None or len(values['URI']) == 0) :
            raise ValueError("Please provide a path or URI for the file.")
        if (values['Path'] != None and len(values['Path']) != 0) :
            values['type'] = 'local'
        else :
            values['type'] = 'remote'
        return values

class Freeform(BaseModel):
    class Config:
        extra = 'allow'
    pass

## Multi-Typed tags (data, code, paper)
class LocalBase(StrDescModel):
    Path: DataT[str]
    Destination: OptDataT[str]
    outpath: str =  ""

class RemoteBase(StrDescModel):
    URI: DataT[AnyUrl]
    Filename: DataT[str]
    outpath: str = ""

class D_Ref(RemoteBase) :
    Type: OptDataT[str]

class D_Copy(LocalBase) :
    Type: OptDataT[str]

class D_Primary(LocalBase) :
    Type: OptDataT[str]

class Data(BaseModel) :
    Ref: OptDataT[D_Ref]
    Copy: OptDataT[D_Copy]
    Primary: OptDataT[D_Primary]

class P_Ref(RemoteBase) :
    Link: OptDataT[AnyUrl]
    PMID: OptDataT[int]
    #Add a validator for PMID?
    DOI: OptDataT[datetime.date]

class P_Copy(StrDescModel) :
    Link: OptDataT[AnyUrl]
    PMID: OptDataT[int]
    #Add a validator for PMID?
    DOI: OptDataT[datetime.date]

class P_Primary(StrDescModel) :
    Link: OptDataT[AnyUrl]
    PMID: OptDataT[int]
    #Add a validator for PMID?
    DOI: OptDataT[datetime.date]

class Paper(BaseModel) :
    Ref: OptDataT[P_Ref]
    Copy: OptDataT[P_Copy]
    Primary: OptDataT[P_Primary]

class S_Ref(RemoteBase):
    Type: DataT[str]
    Version: OptDataT[str]
    
class S_Copy(LocalBase):
    Type: DataT[str]
    Version: OptDataT[str]

class S_Primary(LocalBase):
    Type: DataT[str]
    Version: OptDataT[str]

class Software(BaseModel): 
    Ref: OptDataT[S_Ref]
    Copy: OptDataT[S_Copy]
    Primary: OptDataT[S_Primary]

################################
# Overarching Model            #
################################
# Meant to store every single possible tag that we have defined
class Entity(BaseModel):
    MEDFORD: DataT[MEDFORDmodel]
    Paper: OptDataT[Paper]
    Journal: OptDataT[Journal]
    Date: OptDataT[Date]
    Contributor: OptDataT[Contributor]
    Funding: OptDataT[Funding]
    Keyword: OptDataT[Keyword]
    Species: OptDataT[Species]
    Method: OptDataT[Method]
    Software: OptDataT[Software]
    Data: OptDataT[Data]
    File: OptDataT[ArbitraryFile]
    Freeform: OptDataT[Freeform]

# Temporarily set to BaseModel instead of Entity for testing purposes.
class BCODMO(BaseModel):
    Data: List[Data]
    Contributor: List[Contributor]
    Project: List[Project]
    Expedition: List[Expedition]


import re
from typing import Tuple, Union

from MEDFORD.medford_error_mngr import error_mngr, mfd_duplicated_macro, mfd_remaining_template, mfd_unexpected_macro, mfd_no_desc, mfd_wrong_macro_token

class detail_return():
    type: str
    is_novel : bool = False
    detail : 'detail' = None
    macro : Tuple[str, Tuple[int, str]] = None

    def __init__(self, type, is_novel, detail, macro):
        self.type = type
        if type == "detail_return" :
            self.is_novel = is_novel
            self.detail = detail
        elif type == "macro_return" :
            self.macro = macro
        else : 
            raise Exception("Unexpected behavior in detail.FromLine. Attempted to create detail_return with type '{}'".format(type))

class detail() :
    #static
    template_flag = "[..]"
    macro_head = "`@"
    macro_flag = "`@"
    comment_head = "#"
    comment_flag = "# "
    detail_head = "@"

    macro_dictionary = {}

    #instance
    Major_Tokens = []
    Combined_Major_Token = ""
    Minor_Token = ""
    Line_Number = -1
    Depth = -1
    Data = ""

    def __init__(self, majors, minor, lineno, depth, data):
        self.Major_Tokens = majors
        self.Combined_Major_Token = '_'.join(majors)
        self.Minor_Token = minor
        self.Line_Number = lineno
        self.Depth = depth
        self.Data = data

    @classmethod
    def _clear_cache(cls) :
        detail.macro_dictionary = {}

    @classmethod
    def _validate_comment(cls, **params) :
        pass

    @classmethod
    def _remove_inline_comment(cls, line:str) :
        if detail.comment_flag in line:
            return line.split(detail.comment_flag)[0].strip()

    @classmethod
    def _validate_noncomment(cls, line:str, lineno:int, previous_return: Union[None, detail_return, str], err_mngr:error_mngr) :
        if detail.template_flag in line:
            err_mngr.add_syntax_err(mfd_remaining_template(lineno))

    @classmethod
    def _handle_macro_definition(cls, line:str, lineno:int, err_mngr:error_mngr) :
        # TODO: Check that they haven't put an extra space between `@ and the macro name?
        macro_name, macro_body = line.split(detail.macro_flag,1)[1].split(" ",1)
        
        if macro_name in detail.macro_dictionary.keys() :
            old_lineno = detail.macro_dictionary[macro_name][0]
            err_mngr.add_syntax_err(mfd_duplicated_macro(lineno, old_lineno, macro_name))
        
        # If this is duplicated, we're not going to finish parsing anyways. 
        # Just set the macro value to be the new value for now.
        macro_data = (lineno, macro_body)
        detail.macro_dictionary[macro_name] = macro_data
        return detail_return('macro_return', None, None, (macro_name, macro_data))

    @classmethod
    def _substitute_macro(cls, data:str, lineno:int, err_mngr:error_mngr) :
        curled_macro_regex = "{}\{{[a-zA-Z0-9_]+\}}".format(detail.macro_flag)
        if re.search(curled_macro_regex, data) :
            # we've found a macro that's curled
            macro_loc = re.search(curled_macro_regex, data).span()
            macro_fullname = data[macro_loc[0]:macro_loc[1]] # includes macro head and curly brackets

            macro_name_start = macro_loc[0] + len(detail.macro_flag) + 1 # +1 because of curly bracket
            macro_name_end = macro_loc[1] - 1 # -1 because of curly bracket
            found_macro_name = data[macro_name_start:macro_name_end]
        else :
            # we've found a macro that's not curled
            macro_loc = re.search("{}[a-zA-Z0-9_]\w+".format(detail.macro_flag), data).span() # TODO : check that this works too :clown:
            macro_fullname = data[macro_loc[0]:macro_loc[1]] # includes macro head

            macro_name_start = macro_loc[0] + len(detail.macro_head) # has to move head pointer b/c of macro
            macro_name_end = macro_loc[1]
            found_macro_name = data[macro_name_start:macro_name_end]
            
        if found_macro_name in detail.macro_dictionary.keys() :
            data = data.replace(macro_fullname, detail.macro_dictionary[found_macro_name][1])
        else :
            err_mngr.add_syntax_err(mfd_unexpected_macro(lineno, found_macro_name))
        return data

    @classmethod
    def FromLine(cls, line: str, lineno: int, previous_return: Union[None, detail_return], err_mngr: error_mngr) -> Union[None, detail_return] :
        """Generate a Detail object from a line, the line number, and the Detail generated directly previous.

        :param line: A string to parse into a detail, from a MFD file.
        :type line: str
        :param lineno: The line number in the MFD file that 'line' came from.
        :type lineno: int
        :param prev_detail: The previous detail generated from this MEDFORD file, if applicable.
        :type prev_detail: Detail|None
        :raises ValueError: A Template marker ("[...]") is observed in the provided line.
        :raises MedfordFormatException: A Macro name that has already been used is detected.
        :raises MedfordFormatException: A Macro is referred to that has not yet been defined.
        :raises ValueError: A catch-all error when we have no idea how to process the line provided.
        :return: A 3-tuple describing : Is this detail data, is this NOVEL detail data, and the Detail object. The 'novel' data flag was added to accomodate multi-line details.
        :rtype: Tuple(Bool, Bool, Detail)
        """
        line = line.strip()
        
        # Line contains a comment
        if detail.comment_flag in line :
            line = detail._remove_inline_comment(line)

        # Line is empty
        if(len(line) == 0) :
            return previous_return
        
        # Line IS a comment
        if(line[:len(detail.comment_head)] == detail.comment_head) :
            return previous_return

        # Generic validation for anything that isn't a comment; raises if something is wrong
        # but otherwise continues.
        #   (e.g. a template marker is still present)
        detail._validate_noncomment(line, lineno, previous_return, err_mngr)

        # Line is defining a macro
        if line[:len(detail.macro_head)] == detail.macro_head :
            return detail._handle_macro_definition(line, lineno, err_mngr)

        elif line[:len(detail.macro_head)] == "'@" :
            err_mngr.add_syntax_err(mfd_wrong_macro_token(lineno)) 
            # If they tried & failed to define a macro, should reset previous return.
            return None

        # Line follows the standard major-minor format
        elif line[0] == "@" :
            tokens, body = str.split(line, " ", 1)
            tokens = str.replace(tokens, '@', "")
            tokens_list = tokens.split("-")
            major_tokens = tokens_list[0].split("_")

            if len(tokens_list) == 1 :
                minor_token = "desc"
            else :
                if not previous_return.type == "detail_return" :
                    err_mngr.add_syntax_err(mfd_no_desc(lineno, tokens))
                else :
                    if not previous_return.detail.Major_Tokens == major_tokens :
                        err_mngr.add_syntax_err(mfd_no_desc(lineno, tokens))
                minor_token = tokens_list[1]
            depth = len(tokens_list)
            data = body

            if detail.macro_flag in data :
                data = detail._substitute_macro(data, lineno, err_mngr)

            return detail_return("detail_return", True, cls(major_tokens, minor_token, lineno, depth, data), None)

        # Line is a run-over from a previous line.
        else :
            if previous_return.type == "detail_return" :
                previous_return.detail.addData(line)
                return detail_return("detail_return", False, previous_return.detail, None)

            elif previous_return.type == "macro_return" :
                detail.addMacroData(str(previous_return), line)
                # Not technically necessary -- can just return previous return.
                #   changing it though to make it match the detail_return case for legibility
                return detail_return("macro_return", None, None, previous_return.macro)

            else :
                # I think this will only occur if someone basically just starts typing in an empty file, with no
                #   macro or major-minor tokens?
                raise Exception("Unhandled previous_return type: " + str(previous_return.type) + " in detail.FromLine")
                #err_mngr.add_major_parsing_error(lineno, "formatting", f"Line " + str(lineno) + \
                #    " does not lead with a @ or a # (has neither a token nor is a comment.) " + \
                #    "Did you forget to declare a token? Or did you mean to make this a comment?")


    # TODO: add logic to check if next line leads with space or not.
    def addData(self, line) :
        self.Data = self.Data + " " + line
    
    @classmethod
    def addMacroData(cls, macro_name, line) :
        detail.macro_dictionary[macro_name] = detail.macro_dictionary[macro_name] + " " + line

    def tabstring(self) :
        if self.Minor_Token != 'desc' :
            return('\t'*self.depth + self.Minor_Token + ': ' + self.Data)
        else :
            return(self.Minor_Token + ': ' + self.Data)


from MEDFORD.medford_models import *
from shutil import copyfile, make_archive
from typing import Callable, List, Optional, Tuple, Union, Iterable
from MEDFORD.medford_detailparser import detailparser
import hashlib
import itertools
import os
from copy import Error, deepcopy

## Helper Functions
### Hashing
def calculate_sha_512(filename) :
    with open(filename, "rb") as f:
        bytes = f.read()
        hash = hashlib.sha512(bytes).hexdigest()
    return hash

def calculate_sha_256(filename) :
    with open(filename, "rb") as f:
        bytes = f.read()
        hash = hashlib.sha256(bytes).hexdigest()
    return hash

### Settings
class bagit_settings:
    prefix = "data/"
    regex = "^[^%\s]+$"
    _bagdir = ""
    _datadir = ""

    @classmethod
    def set_datadir(cls, input_dir:str) -> None :
        bagit_settings._datadir = input_dir
    
    @classmethod
    def get_datadir(cls) -> str :
        if cls._datadir == "" :
            raise Error("Data directory not set")
        return cls._datadir

    @classmethod
    def set_bagdir(cls, input_dir:str) -> None :
        cls._bagdir = input_dir
    
    @classmethod
    def get_bagdir(cls) -> str :
        if cls._bagdir == "" :
            raise Error("Bag directory not set")
        return cls._bagdir

LocalFile = Union[ArbitraryFile, D_Copy, D_Primary, S_Copy, S_Primary]
RemoteFile = Union[ArbitraryFile, D_Ref, S_Ref]

### Complex Helpers
def add_outpath(inp_file: Union[LocalFile, RemoteFile], type:str) -> Union[LocalFile, RemoteFile]:
    """Given a either a remote or local file, defines its new bag-relative location.
    
    INPUT:
        inp_file (r)    : the pydantic model of the file's information. Must have a Name. Subdirectory may or may not be defined.

    OUTPUT:
        LocalFile | RemoteFile - the same file, but with a new outpath defined.
    """
    datadir = bagit_settings.get_datadir()
    if type == "local":
        # 1. create the novel location
        if inp_file.Destination is not None :
            # TODO: separate this out into a separate function. (filename mngmt utils?)
            if inp_file.Destination[0][1][0] == "/" :
                shortdest = inp_file.Destination[0][1][2:]
            else :
                shortdest = inp_file.Destination[0][1]
            inp_file.outpath = [(inp_file.Destination[0][0], datadir + shortdest)]
        else :
            inp_file.outpath = [(-1, datadir + os.path.basename(inp_file.Path[0][1]))]
    elif type == "remote":
        inp_file.outpath = [(inp_file.Filename[0][0], "data/" + inp_file.Filename[0][1])]
    
    return inp_file

def copy_local_file(inp_file: str, copy_location: str) -> Tuple[bool, Union[None, Error]]:
    """
    Given a "local" file (file with a defined PATH), copies it from its original path to its bag-relative location.
    
    INPUT: 
        **inp_file** (required) : str representing the current location of the file in the operating system
        **copy_location** (required) : the location that was determined to be the bag-relative location this file should be copied to.
    
    OUTPUT:
        Tuple[bool, None | Error] - a tuple where the first bool represents whether or not the copy was successful. If the copy was unsuccessful (bool is False), the error message is stored in the second entry of the Tuple.

    SIDE EFFECTS:
        - Copies the local file from Path to _output_path.
    """
    try:
        copy_directory = os.path.dirname(copy_location)
        os.makedirs(copy_directory, exist_ok=True)
        copyfile(inp_file, copy_location)
        return((True, None))
    except OSError as e:
        return((False, e))

def mutate_local_file(inp_file: LocalFile) -> None :
    """
    Given a local file and its relative location, adjusts its pydantic model to represent the new bag-relative location.

    INPUT:
        **inp_file** (required) : the pydantic model of the file's information.
    
    SIDE EFFECTS:
        - changes Path in input_file to point to relative_location + inp_file.Name[0], and deletes Subdirectory if defined.
    """
    
    inp_file.Path = inp_file.outpath
    inp_file.outpath = None
    inp_file.Destination = None
    if 'type' in inp_file.__dict__.keys() :
        del inp_file.type
    return inp_file

def mutate_remote_file(inp_file: RemoteFile) -> None:
    inp_file.Filename = inp_file.outpath
    inp_file.outpath = None
    if 'type' in inp_file.__dict__.keys() :
        del inp_file.type
    return inp_file


def manage_remote_file(inp_file: RemoteFile) -> str:
    """Creates the fetch.txt line for a file

    Given a "remote file", generates the line to be entered into fetch.txt.

    INPUT:
        inp_file (r)    : the pydantic model of the file's information. Must have a Name and a URI.
    
    OUTPUT:
        str     - a string that fulfills all the requirements for fetchtxt, formatted as follows:
            (inp_file URI)  -   (inp_file output location)
    """
    outpath = inp_file.outpath[0][1]

    fetchstring = inp_file.URI[0][1] + "\t-\t" + outpath
    return fetchstring

def perform_medford_munging(mfd_input: str, howto_write: Callable[[ArbitraryFile, str],None]) -> Tuple[str, int] :
    """Create an entry for the current medford file, write the adjusted file into the bag.

    Given the current medford file location, creates an ArbitraryFile entry for the medford. Then, calculates what its location inside the bag should be and modifies the ArbitraryFile entry appropriately. Then uses the passed howto_write function to write the fully adjusted medford file to the correct location before finally returning all the information required for the manifest.

    INPUT:
        mfd_input (r)   : (str) The location of the .mfd file that MEDFORD is currently running on.
        howto_write (r) : (Callable[[ArbitraryFile, str], None]) A function that given the medford pydantic model and an output location, writes the adjusted medford file in medford format to that location.
        
    OUTPUT:
        Tuple[str, int] : A tuple containing all of the information required for the manifest (bag-relative location and sha integer.))
    """
    # TODO: foolproof way of getting medford file name
    mfd_model = ArbitraryFile(desc = [(-1,'Medford File')], Path = [(-1,mfd_input)])
    mfd_model = add_outpath(mfd_model, 'local')
    outpath = mfd_model.outpath[0][1] # store the outpath rq (not technically necessary)

    mfd_model = mutate_local_file(mfd_model) # remember: this mutates!!! (how do I flag that? NoneType return? Should I make this not mutate but return a new object with the values changed? Probably...)
    howto_write(mfd_model, outpath)

    sha = calculate_sha_512(mfd_model.Path[0][1])
    return (mfd_model.Path[0][1], sha)

def write_manifest(sha_entries) :
    outdir = bagit_settings.get_bagdir()
    with open(outdir + "manifest-sha512.txt", 'w') as f:
        for entry in sha_entries:
            f.write(entry)
            f.write("\n")

def write_fetch(fetch_entries) :
    outdir = bagit_settings.get_bagdir()
    with open(outdir + "fetch.txt", 'w') as f:
        for entry in fetch_entries :
            f.write(entry)
            f.write("\n")

def zip_all_files(mfd_input) :
    outdir = bagit_settings.get_bagdir()
    # WARNING: VERY fragile; not tested on windows or macs whatsoever.
    zipname = os.path.splitext(os.path.basename(mfd_input))[0]
    dirname = os.path.dirname(mfd_input)
    make_archive(dirname + "/" + zipname + ".zip", "zip", outdir)

## Models
class BagIt(Entity) :
    Data: OptDataT[Data]

## Runners
def runBagitMode(parameters, medford_input) :
    # Goals for BagIt processing:
    #   1. generate the bagIt output directory string
    #   2. create empty fetch.txt file
    #   3. create empty manifest-sha512.txt file
    #   4. Identify all the files to be included in the bag
    #       a. arbitrary data files :
    #            i. with a PATH defined
    #            ii. with a URI defined
    #       b. Data files:
    #            i. copy            : has a PATH, might have a URI      : copy the file from the PATH to the bag-relative location & add to sha
    #            ii. primary        : has a PATH, shouldn't have a URI  : copy the file from the PATH to the bag-relative location & add to sha
    #            iii. reference     : has a URI, shouldn't have a PATH  : add to the fetch.txt file
    #       c. Software files:
    #          i. copy            : has a PATH, might have a URI      : copy the file from the PATH to the bag-relative location & add to sha
    #          ii. primary        : has a PATH, shouldn't have a URI  : copy the file from the PATH to the bag-relative location & add to sha
    #          iii. reference     : has a URI, shouldn't have a PATH  : add to the fetch.txt file
    #   5. Modify MEDFORD file s.t. PATHs now all point to novel locations
    #   6. Copy MEDFORD into the bag
    #   7. add MEDFORD file to manifest-sha512.txt
    #   8. zip entire bag


    # generating the output directory strong
    bagdir = medford_input + "_BAG/"
    if not os.path.isdir(bagdir) :
        os.mkdir(bagdir)
    bagit_settings.set_bagdir(bagdir)

    bagdatadir = bagdir + "data/"
    if not os.path.isdir(bagdatadir) :
        os.mkdir(bagdatadir)
    bagit_settings.set_datadir(bagdatadir)

    # Deep copying parameters because we are about to do a LOT of mutation
    _parameters = deepcopy(parameters)
    bagit_settings.output_location = bagdir

    # create empty fetch.txt file

    fetch_lines = []

    # create empty manifest-sha512.txt file
    sha_lines = []

    # Identify all the files to be included in the bag
    local_files = []
    remote_files = []
    
    # arbitrary files
    if _parameters.File is not None and len(_parameters.File) > 0:
        for possible_file in _parameters.File :
            if possible_file[1].type == 'local' :
                local_files.append(possible_file)
            elif possible_file[1].type == 'remote' :
                remote_files.append(possible_file)
            else :
                raise NotImplementedError #fix error type

    # data & software files
    local_named = []
    remote_named = []
    if _parameters.Data is not None :
        local_named = list(itertools.chain.from_iterable(itertools.dropwhile(lambda x: x is None, [_parameters.Data[0][1].Copy, _parameters.Data[0][1].Primary])))
        remote_named = _parameters.Data[0][1].Ref if _parameters.Data[0][1].Ref is not None else []
    if _parameters.Software is not None :
        local_named = local_named + list(itertools.chain.from_iterable(itertools.dropwhile(lambda x: x is None, [_parameters.Software[0][1].Copy, _parameters.Software[0][1].Primary])))
        remote_named = remote_named + _parameters.Software[0][1].Ref if _parameters.Software[0][1].Ref is not None else []

    all_local_files = itertools.chain(local_named, local_files)
    modified_locals = []
    for lf in all_local_files :
        lf = lf[1]
        lf = add_outpath(lf, 'local')
        curpath = lf.Path[0][1]
        curout = lf.outpath[0][1]
        
        # 2. create the sha & add to manifest
        lf_sha = calculate_sha_512(curpath)
        sha_lines.append("%s %s" % (curpath, lf_sha))

        # 3. copy the file to the novel location
        copy_local_file(curpath, curout)

        # 4. update the internal information (path -> new location)
        lf = mutate_local_file(lf)
        
        # store the modified local file
        modified_locals.append(lf)
        
    all_remote_files = itertools.chain(remote_named, remote_files)
    for rf in all_remote_files :
        rf = rf[1]
        rf = add_outpath(rf, 'remote')

        rf_fetch = manage_remote_file(rf)
        rf = mutate_remote_file(rf)
        fetch_lines.append(rf_fetch)

    def write_medford_file(mfd_model, final_location) :
        _parameters.File.append((-1,mfd_model))
        detailparser.write_from_model(_parameters, final_location)

    mfd_path, mfd_sha = perform_medford_munging(medford_input, write_medford_file)
    sha_lines.append("%s %s" % (mfd_path, mfd_sha))
    
    write_manifest(sha_lines)
    write_fetch(fetch_lines)

    zip_all_files(medford_input)




# Take 3!
from copy import deepcopy
import functools
import math

from pydantic import BaseModel

from MEDFORD.medford_detail import *
from MEDFORD.medford_error_mngr import *

class detailparser :

    def __init__(self, details, err_mngr) :
        self.data = {}
        self.parse_details(details)
        self.err_mngr = err_mngr

    def parse_details(self, details) :
        prev_major = "-1"
        for detail in details :
            cur_minor = detail.Minor_Token
            cur_major = detail.Combined_Major_Token
            if not cur_major == prev_major:
                if not detail.Minor_Token == "desc" :
                    error_mngr.add_syntax_err(mfd_no_desc(detail.Line_Number, cur_major))
                    # Add a dummy desc so we can keep running for now.
                    self.add_to_dict(detail.Major_Tokens, "desc", "NO DESC PROVIDED", detail.Line_Number, False)
                    self.add_to_dict(detail.Major_Tokens, detail.Minor_Token, detail.Data, detail.Line_Number, False)
                else :
                    self.add_to_dict(detail.Major_Tokens, detail.Minor_Token, detail.Data, detail.Line_Number, True)

            elif cur_minor == "desc" :
                self.add_to_dict(detail.Major_Tokens, detail.Minor_Token, detail.Data, detail.Line_Number, True)

            else :
                self.add_to_dict(detail.Major_Tokens, detail.Minor_Token, detail.Data, detail.Line_Number, False)
            prev_major = cur_major

    def add_to_dict(self, major_list, minor, data, lineno, append=False):
        cur_dict = self.data
        for n,major in enumerate(major_list) :
            if major not in cur_dict.keys() :
                cur_dict[major] = [(lineno, {})]
                cur_dict = cur_dict[major][0][1]
            else :
                if append == True and n == len(major_list) - 1:
                    cur_dict[major].append((lineno, {}))
                cur_dict = cur_dict[major][-1][1]

        if minor not in cur_dict.keys() :
            cur_dict[minor] = [(lineno, data)]
        else :
            cur_dict[minor].append((lineno, data))

    def export(self) :
        return deepcopy(self.data)

    def travel_major_tokens(curdat:BaseModel, current_majors: List[str]) -> List[str] :
        set_attributes = curdat.__fields_set__
        all_keys = list(curdat.__fields__.keys())
        collapsed_majors = '_'.join(current_majors)
        output_strings = []
        
        for attr in set_attributes :
            if attr not in all_keys :
                all_keys.append(attr)

        for attr in all_keys :
            if curdat.__getattribute__(attr) != None :
                for val in curdat.__getattribute__(attr) :
                    val = val[1]
                    if issubclass(type(val), BaseModel) :
                        # this is a major token again, so we need to recurse
                        res = detailparser.travel_major_tokens(val, current_majors + [attr])
                        output_strings.extend(res)
                    else :
                        # we're in a minor token, so start writing coach
                        if attr != "desc" :
                                output_strings.append(f'@{collapsed_majors}-{attr} {val}\n')
                        else :
                            output_strings.append(f'@{collapsed_majors} {val}\n')
        
        # if we haven't already had a newline, add a newline
        # ( double-newline happens if we're in a multi-major-token that is not followed by a sister major token,
        #   e.g. :
        #   Data-Primary asdf
        #   Software-Primary fdsa
        # 
        #   instead of:
        #   Data-Primary asdf
        #   Data-Ref fdsa)
        if output_strings[-1] != "\n" :
            output_strings.append("\n")
        return output_strings

    def parse_pydantic_errors(self, errs, dict) :
        errors = errs.errors()
        ## first, compress all related errors
        error_locations = {}
        for e in errors:
            if e['loc'] not in error_locations.keys() :
                error_locations[e['loc']] = [e]
            else :
                error_locations[e['loc']].append(e)

        for error_loc in error_locations.keys() :
            error_type = None
            current_errors = error_locations[error_loc]
            # are all of these just the result of there being multiple valid types, and none of them were found?
            if len(current_errors) > 1 :
                all_type_errors = functools.reduce(lambda a, b: a and ("value_error" in b['type']) and (not "missing" in b['type']), current_errors)
                if all_type_errors == True :
                    # FOR NOW just take the first one
                    error_msg = current_errors[0]['msg']
                    error_type = "value_error"
                else :
                    raise NotImplementedError("Multiple error types have been raised at the same location. Check this out!")
            else :
                if "missing" in current_errors[0]['type'] :
                    error_type = "missing_field"
                elif "incomplete_data_error" in current_errors[0]['type']:
                    error_type = "incomplete_data"
                elif "value_error" in current_errors[0]['type'] :
                    error_type = "value_error"
                elif "type_error" in current_errors[0]['type'] :
                    error_type = "value_error" # hurts to type, but...
                error_msg = current_errors[0]['msg']

            # Below this, we assume that we only care about the first error.

            ##### LINE NUMBER #####
            # Get the index in pydantic's 'loc' n-ple that corresponds to the line number of the most specific relevant block.
            n_3ples = math.floor(len(error_loc)/3.0)
            start_last3 = (n_3ples-1)*3
            end_last3 = (n_3ples)*3
            location_hint_slice = error_loc[:(end_last3-1)]

            # Get the line number that index points to.
            t_err = dict
            for location_hint in location_hint_slice :
                t_err = t_err[location_hint]
            if len(location_hint_slice) == 0 :
                error_line_number = -1
            else:
                error_line_number = t_err[0]

            ##### RELATED TOKENS #####
            if n_3ples > 0:
                token_indices = list(range(0, (n_3ples)*3,3))
            else :
                token_indices = [0]
            tokens = [error_loc[i] for i in token_indices]
            token_string = "@" + tokens[0]
            if len(tokens) > 1:
                for token in tokens[1:-1]:
                    token_string = token_string + "_" + token
                token_string = token_string + "-" + tokens[-1]

            error_obj = mfd_err(error_line_number, error_type, tokens, error_loc[-1], error_msg)
            self.err_mngr.add_error(error_obj)
        self.err_mngr.print_errors()


    def write_from_model(model:BaseModel, location:str) -> None:
        with open(location, 'w') as f:
            lines = detailparser.travel_major_tokens(model, [])
            # Because I wrote it, I know lines is going to end up with an extra newline.
            # Remove it for cleanliness.
            if lines[-1] == "\n" :
                lines = lines[:-1]
                # Again, because I wrote it, I know that every line ends with a newline.
                # Remove the newline from the last line so we don't have a trailing newline.
                lines[-1] = lines[-1][:-1]
            for line in lines :
                f.write(line)



from pydantic import BaseModel, ValidationError
from MEDFORD.medford_detailparser import *
from MEDFORD.medford_detail import *
from MEDFORD.medford_models import BCODMO, Entity, StrDescModel
from MEDFORD.medford_BagIt import runBagitMode, BagIt
from MEDFORD.medford_error_mngr import *
import json

import argparse

from enum import Enum, auto


class MFDMode(Enum) :
    OTHER = 'OTHER'
    BCODMO = 'BCODMO'
    BAGIT = 'BAGIT'

    def __str__(self):
        return self.value

class ParserMode(Enum) :
    validate = 'validate'
    compile = 'compile'

    def __str__(self):
        return self.value

class ErrorMode(Enum) :
    first = 'FIRST'
    all = 'ALL'

    def __str__(self) :
        return self.value

class ErrorOrder(Enum) :
    type = 'TYPE'
    tokens = 'TOKENS'
    line = 'LINE'

    def __str__(self):
        return self.value


# Command Line Arguments
# TODO: How can I make it require an MFDMode value, but lowercase is OK?
#       Do I actually care?
parser = argparse.ArgumentParser()
parser.add_argument("-m", "--mode", type=MFDMode, choices=list(MFDMode), default=MFDMode.OTHER, required=True,
    help="Which Output mode the MEDFORD parser should validate or compile for.")
parser.add_argument("action", type=ParserMode, choices=list(ParserMode),
    help="Whether the MEDFORD parser is only validating or actually compiling (performing any necessary adjustments or actions for the appropriate format, such as creating a Bag for the BagIt mode.)")
parser.add_argument("file", type=str, help="The input MEDFORD file to validate or compile.")
parser.add_argument("--write_json", action="store_true", default=False,
    help="Write a JSON file of the internal representation of the MEDFORD file beside the input MEDFORD file.")
parser.add_argument("--debug", "-d", "-v", action="store_true", default=False,
    help="Enable verbose mode for MEDFORD, enabling various debug messages during runtime.")
parser.add_argument("--error_mode", "-e", type=ErrorMode, choices=list(ErrorMode), default=ErrorMode.all,
    help="(ALL | FIRST) Whether to compile all errors or stop on the first error encountered.")
parser.add_argument("--error_sort", "-s", type=ErrorOrder, choices=list(ErrorOrder), default=ErrorOrder.line,
    help="(TYPE|TOKENS|LINE) How to sort the errors, if compiling all errors.")

def runMedford(filename, output_json, mode, error_mode, error_sort, action):
    class FieldError(Exception):
        pass

    details = []
    err_mngr = error_mngr(str(error_mode), str(error_sort))

    # TODO: add error catching for mis-formatting in here...
    # to test, change line 12 of pshpil_rnaseq.mfd to have a typo in the macro name.
    with open(filename, 'r') as f:
        all_lines = f.readlines()
        dr = None
        for i, line in enumerate(all_lines):
            if(line.strip() != "") :
                # TODO: move the details collection logic to detail? I don't like that we have to pull the typing here.
                dr = detail.FromLine(line, i+1, dr, err_mngr)
                if isinstance(dr, detail_return) :
                    if dr.is_novel :
                        details.append(dr.detail)
    
    if err_mngr.has_major_parsing :
        err_mngr.print_syntax_errors()
        raise SystemExit(0)

    parser = detailparser(details, err_mngr)
    final_dict = parser.export()
    p = {}
    # nom
    try:
        if mode == MFDMode.BCODMO:
            p = BCODMO(**final_dict)
        elif mode == MFDMode.BAGIT:
            p = BagIt(**final_dict)
        elif mode == MFDMode.OTHER:
            p = Entity(**final_dict)
        else :
            raise Exception("Medford is running in an unsupported mode.")
    except ValidationError as e:
        parser.parse_pydantic_errors(e, final_dict)
    else:
        print("No errors found.")

    if mode == MFDMode.BAGIT and action == ParserMode.compile:
        runBagitMode(p, filename)

    if(output_json) :
        with open(filename + ".JSON", 'w') as f:
            json.dump(final_dict, f, indent=2)

def parse_args_and_go() :
    args = parser.parse_args()
    runMedford(args.file, args.write_json, args.mode, args.error_mode, args.error_sort, args.action)

if __name__ == "__main__":
    parse_args_and_go()
